//-----------------------------------------------------------------------------
// MK3_2560_SERVO
// Version 1.1
// 02.11.2011 / UB
// "www.Mikrocontroller-4U.de"
// mc-4u(@)t-online.de
//-----------------------------------------------------------------------------
// Fuer das myAVR MK3 Board
// mit ATMega-2560 @ 16MHz
// Modellbau-Servo Ansteuerung (Aufloesung = 10Bit)
// (bis zu 8 Servos gleichzeitig)
//-----------------------------------------------------------------------------
// das File muss im Ordner
// "myAVR/WinAVR/AVR/Include/UB_MK3_2560/"
// liegen und mit #include <UB_MK3_2560/MK3_2560_SERVO.h>
// eingebunden werden
//-----------------------------------------------------------------------------
// Belegt : maximal 8 Port-Bits (default = Port-L Bit0 bis Bit3)
//          Timer-1 (16bit,Mode4=CTC)
//-----------------------------------------------------------------------------

#include <avr/interrupt.h>


//-----------------------------------------------------------------------------
// INFOS zur Benutzung :
// 01. Funktion "MK3_INIT_SERVO" starten
// 02. Funktion "MK3_SERVO_SET" zum bewegen eines Servos
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Enumerationen fuer die Funktioen
//-----------------------------------------------------------------------------


typedef enum {
  SERVO_1 = 0,
  SERVO_2,
  SERVO_3,
  SERVO_4,
  SERVO_5,
  SERVO_6,
  SERVO_7,
  SERVO_8
}SERVO_NR;



//-----------------------------------------------------------------------------
// Defines
//-----------------------------------------------------------------------------
#define MIN_SERVO_POS 1000      // Min_Impulswert (in us) -> linker Anschlag
#define MAX_SERVO_POS 2000      // Max_Impulswert (in us) -> rechter Anschlag

#define MAX_SERVO_WERT 1023     // Servo Maxwert (fuer 10Bit = 1023)

#define SERVO_DELTA (MAX_SERVO_POS-MIN_SERVO_POS)  // Servo Bewegungsbereich
#define SERVO_STEP  (SERVO_DELTA/1023.0)           // Servo Aufloesung (10Bit)




//-----------------------------------------------------------------------------
// Funktionen die vom MAIN benutzt werden koennen
//-----------------------------------------------------------------------------
void MK3_INIT_SERVO(void);	                   // zum initialisieren der Servos
void MK3_SERVO_SET(SERVO_NR nr,unsigned int pos);  // zum bewegen eines Servos
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
// Port-Pin definition
//-----------------------------------------------------------------------------
#define MK3_DDR_SERVO DDRL	// DDR-Register der Servos
#define MK3_PORT_SERVO PORTL	// Port der Servos

//-----------------------------------------------------------------------------
// hier muessen die Port-Pins der benutzten Servos eingetragen werden
// Wichtig !! falls ein Servo nicht benutzt wird -> 0x00 eintragen !!
//-----------------------------------------------------------------------------

#define MK3_SERVO_1 0x01        // PortPin von Servo Nr.1
#define MK3_SERVO_2 0x02        // PortPin von Servo Nr.2
#define MK3_SERVO_3 0x04        // PortPin von Servo Nr.3
#define MK3_SERVO_4 0x08        // PortPin von Servo Nr.4
#define MK3_SERVO_5 0x00        // PortPin von Servo Nr.5 (nicht benutzt)
#define MK3_SERVO_6 0x00        // PortPin von Servo Nr.6 (nicht benutzt)
#define MK3_SERVO_7 0x00        // PortPin von Servo Nr.7 (nicht benutzt)
#define MK3_SERVO_8 0x00        // PortPin von Servo Nr.8 (nicht benutzt)



//-----------------------------------------------------------------------------
// Interne Variabeln und Funktionen
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Struktur fuer einen Servo-Kanal
//-----------------------------------------------------------------------------
typedef struct {
  unsigned int impuls; // Servo-Impulswert
  unsigned int timer;  // Timer-Wert
  unsigned char pin;   // Servo-Port-Pin  
}SServo_t;

//-----------------------------------------------------------------------------
// Struktur fuer alle Servos
//-----------------------------------------------------------------------------
typedef struct {
  SServo_t nr[9];       // alle 8 Servos + 1 Dummy Servo
  unsigned char akt_nr; // aktuelle Servo-Nr 
}MServo_t;
MServo_t SERVO;


//-----------------------------------------------------------------------------
void MK3_INIT_SERVO(void) {	// zum initialisieren der Servos
  // alle Servos initialisieren

  // alle Servo-Port-Pins als Ausgang definieren
  MK3_DDR_SERVO |= MK3_SERVO_1;
  MK3_DDR_SERVO |= MK3_SERVO_2;
  MK3_DDR_SERVO |= MK3_SERVO_3;
  MK3_DDR_SERVO |= MK3_SERVO_4;
  MK3_DDR_SERVO |= MK3_SERVO_5;
  MK3_DDR_SERVO |= MK3_SERVO_6;
  MK3_DDR_SERVO |= MK3_SERVO_7;
  MK3_DDR_SERVO |= MK3_SERVO_8;

  // alle Servo-Port-Pins auf LO-Pegel
  MK3_PORT_SERVO &= ~MK3_SERVO_1;
  MK3_PORT_SERVO &= ~MK3_SERVO_2;
  MK3_PORT_SERVO &= ~MK3_SERVO_3;
  MK3_PORT_SERVO &= ~MK3_SERVO_4;
  MK3_PORT_SERVO &= ~MK3_SERVO_5;
  MK3_PORT_SERVO &= ~MK3_SERVO_6;
  MK3_PORT_SERVO &= ~MK3_SERVO_7;
  MK3_PORT_SERVO &= ~MK3_SERVO_8;

  //----------------------------------------------
  // Timer fuer den Counter : Timer-1
  // im Mode4 -> ClearTimerOnCompare-A Match
  // Vorteiler = 8
  // CompA Wert = Periodendauer (20ms)
  // CompB Wert = Impulswert (variabel 1ms bis 2ms)
  //
  // VORSICHT !! -> diese Einstellungen nicht aendern
  //----------------------------------------------

  TCCR1A=0x00;         // normal Port operation
  TCCR1B=0x08;         // Mode4, CTC
  TCCR1C=0x00;

  TCNT1=0x0000;        // startwert
  OCR1A=40000;         // Periodendauer (20ms)
  OCR1B=3000;          // Impulswert (1,5ms)

  TIMSK1=0x06;         // Interrupt bei Comp-A + Comp-B   
  
  //----------------------------------------------
  // Init der Servo-Struktur
  // WICHTIG !! -> nicht aendern
  //----------------------------------------------
  SERVO.akt_nr=0;
  SERVO.nr[0].impuls=3000;      // Servo Nr.1
  SERVO.nr[0].timer=3000;
  SERVO.nr[0].pin=MK3_SERVO_1;
  SERVO.nr[1].impuls=3000;      // Servo Nr.2
  SERVO.nr[1].timer=6000;
  SERVO.nr[1].pin=MK3_SERVO_2;
  SERVO.nr[2].impuls=3000;      // Servo Nr.3
  SERVO.nr[2].timer=9000;
  SERVO.nr[2].pin=MK3_SERVO_3;
  SERVO.nr[3].impuls=3000;      // Servo Nr.4
  SERVO.nr[3].timer=12000;
  SERVO.nr[3].pin=MK3_SERVO_4;
  SERVO.nr[4].impuls=3000;      // Servo Nr.5
  SERVO.nr[4].timer=15000;
  SERVO.nr[4].pin=MK3_SERVO_5;
  SERVO.nr[5].impuls=3000;      // Servo Nr.6
  SERVO.nr[5].timer=18000;
  SERVO.nr[5].pin=MK3_SERVO_6;
  SERVO.nr[6].impuls=3000;      // Servo Nr.7
  SERVO.nr[6].timer=21000;
  SERVO.nr[6].pin=MK3_SERVO_7;
  SERVO.nr[7].impuls=3000;      // Servo Nr.8
  SERVO.nr[7].timer=24000;
  SERVO.nr[7].pin=MK3_SERVO_8;
  SERVO.nr[8].impuls=0;         // Dummy Servo
  SERVO.nr[8].timer=50000;      // Wichtiger eintrag !!
  SERVO.nr[8].pin=0x00;         // nicht aendern
  
  //----------------------------------------------
  // alle Servos auf eine Startposition stellen
  // -> hier alle auf linken Anschlag = 1,0ms
  //----------------------------------------------
  MK3_SERVO_SET(SERVO_1,0);
  MK3_SERVO_SET(SERVO_2,0);
  MK3_SERVO_SET(SERVO_3,0);
  MK3_SERVO_SET(SERVO_4,0);
  MK3_SERVO_SET(SERVO_5,0);
  MK3_SERVO_SET(SERVO_6,0);
  MK3_SERVO_SET(SERVO_7,0);
  MK3_SERVO_SET(SERVO_8,0);
  
  //----------------------------------------------
  // Timer starten
  //----------------------------------------------
  TCCR1B|=0x02;        // Timer1 starten (Vorteiler=8)

  SREG|=0x80; // Alle Interrupts einschalten
}

//-----------------------------------------------------------------------------
void MK3_SERVO_SET(SERVO_NR nr,unsigned int pos) { // zum bewegen eines Servos
  // speichert den Impulswert vom aktuellen Servo
  // und aktuallisiert die Timer-Werte aller Folgeservos

  unsigned char n,akt;
  unsigned int int_wert=0;
  float f;
  
  // auf grenzen achten
  if(pos>MAX_SERVO_WERT)
    pos=MAX_SERVO_WERT;  
  
  // Impulswert ausrechnen
  f=(float)(SERVO_STEP)*(float)(pos);
  int_wert=(int)(f)+MIN_SERVO_POS;

  // auf grenzen achten
  if(int_wert<MIN_SERVO_POS)
    int_wert=MIN_SERVO_POS;
  if(int_wert>MAX_SERVO_POS)
    int_wert=MAX_SERVO_POS;
  
  // auf Timer-Wert umrechnen
  int_wert=int_wert*2;
  
  // impulswert vom aktuellen Servo speichern
  // und alle Timer-Werte aller Folgeservos aktuallisieren
  if(nr==SERVO_1) {
    // Servo Nr.1 hat eine Sonderstellung
    SERVO.nr[0].impuls=int_wert;
    SERVO.nr[0].timer=int_wert;
    for(n=1;n<=7;n++) {
      // Timer_Werte der restlichen 7 Servos updaten
      int_wert=SERVO.nr[n-1].timer;
      int_wert+=SERVO.nr[n].impuls;
      SERVO.nr[n].timer=int_wert;
    }
  }
  else {
    // servo suchen
    akt=1;
    if(nr==SERVO_2) akt=1; 
    if(nr==SERVO_3) akt=2;
    if(nr==SERVO_4) akt=3;
    if(nr==SERVO_5) akt=4;
    if(nr==SERVO_6) akt=5;
    if(nr==SERVO_7) akt=6;
    if(nr==SERVO_8) akt=7;
    
    // impulswert speichern
    SERVO.nr[akt].impuls=int_wert;
    
    for(n=akt;n<=7;n++) {
      // Timer_Werte der restlichen Servos updaten
      int_wert=SERVO.nr[n-1].timer;
      int_wert+=SERVO.nr[n].impuls;
      SERVO.nr[n].timer=int_wert;      
    }
  }
}

//-----------------------------------------------------------------------------
// ISR von Timer-1 Compare-A
// wird bei 16MHz Quarz alle 20ms aufgerufen
//-----------------------------------------------------------------------------
ISR (TIMER1_COMPA_vect) {                          // ISR : Timer1 Comp-A

  // auf ersten Servo stellen
  SERVO.akt_nr=0;
  // Comp-B setzen
  OCR1B=SERVO.nr[0].timer;
  // Port-Pin einschalten
  MK3_PORT_SERVO |= SERVO.nr[0].pin;
}

//-----------------------------------------------------------------------------
// ISR von Timer-1 Compare-B
// wird je nach Impulslaenge
// nach 1,0ms bis 2,0ms aufgerufen
//-----------------------------------------------------------------------------
ISR (TIMER1_COMPB_vect) {                          // ISR : Timer1 Comp-B

  // Port-Pin ausschalten
  MK3_PORT_SERVO &= ~SERVO.nr[SERVO.akt_nr].pin;
  // auf naechsten Servo schalten
  SERVO.akt_nr++;
  // Comp-B setzen
  OCR1B=SERVO.nr[SERVO.akt_nr].timer; 
  // Port-Pin einschalten
  MK3_PORT_SERVO |= SERVO.nr[SERVO.akt_nr].pin;  
}








